import Database from "bun:sqlite";
import type { ColumnShorthandMap } from "./columns";
import { SchemaError } from "./error";
import { Table } from "./table";

interface KVRow {
  kv_id: number;
  key: string;
  value: string;
}

/**
 * Represents a key-value store using Prequel.
 */
export class PrequelKVStore {
  private table: Table<KVRow>;

  /**
   * Creates a new instance of the PrequelKVStore class.
   * @param pq The Prequel instance.
   * @param tableName The name of the table.
   */
  constructor(pq: Prequel, tableName: string) {
    this.table = pq.table<KVRow>(tableName, {
      kv_id: {
        type: "INTEGER",
        primary: true,
        autoincrement: true,
      },
      key: {
        type: "TEXT",
        unique: true,
      },
      value: "TEXT",
    });
  }

  /**
   * Inserts or updates a key-value pair in the store.
   * @param key The key.
   * @param value The value.
   */
  put<T>(key: string, value: T): void {
    const lookup = this.table.findOneWhere({
      key,
    });

    if (lookup === null) {
      this.table.insertPartial({
        key,
        value: JSON.stringify(value),
      });
    } else {
      lookup.value = JSON.stringify(value);
      this.table.update(lookup);
    }
  }

  /**
   * Retrieves the value associated with the specified key from the store.
   * @param key The key.
   * @returns The value associated with the key, or null if the key is not found.
   */
  get<T>(key: string): T | null {
    const lookup = this.table.findOneWhere({ key });

    if (lookup === null) {
      return null;
    }

    return JSON.parse(lookup.value) as T;
  }

  /**
   * Checks if a key exists in the table.
   * @param key - The key to check.
   * @returns `true` if the key exists, `false` otherwise.
   */
  has(key: string): boolean {
    return this.table.existsWhere({ key });
  }

  /**
   * Get an array of all of the keys in the KV store
   * @returns The keys in the store.
   */
  keys(): string[] {
    return this.table
      .prepare<{ key: string }, []>(`SELECT key FROM ${this.table}`)
      .all()
      .map(({ key }) => key);
  }

  /**
   * Get an array of all of the values in the KV store
   * @returns The values in the store.
   */
  values(): unknown[] {
    return this.table
      .prepare<{ value: string }, []>(`SELECT value FROM ${this.table}`)
      .all()
      .map(({ value }) => JSON.parse(value));
  }
}

/**
 * Represents a Prequel database instance.
 */
export class Prequel {
  public db: Database;
  public kv: PrequelKVStore;
  public tables: Record<string, Table<unknown>> = {};
  public query: typeof Database.prototype.query;
  public prepare: typeof Database.prototype.prepare;

  /**
   * Creates a new Prequel database instance.
   * @param filename The filename of the database. Defaults to ":memory:" if not provided.
   */
  constructor(filename = ":memory:") {
    this.db = new Database(filename, {
      strict: true,
    });
    this.query = this.db.query.bind(this.db);
    this.prepare = this.db.prepare.bind(this.db);
    this.db.exec("PRAGMA foreign_keys=ON");
    this.kv = new PrequelKVStore(this, "PrequelManagedKVStore");
  }

  /**
   * Enables Write-Ahead Logging (WAL) mode for the database.
   */
  enableWAL() {
    this.db.exec("PRAGMA journal_mode=WAL;");
  }

  /**
   * Closes the database connection.
   */
  close() {
    this.db.close();
  }

  /**
   * Checks if a table with the given name exists in the database.
   * @param name The name of the table.
   * @returns `true` if the table exists, `false` otherwise.
   */
  hasTable(name: string): boolean {
    return Object.keys(this.tables).includes(name);
  }

  /**
   * Creates a new table in the database.
   * @param name The name of the table.
   * @param cols The column definitions of the table.
   * @returns The created table.
   * @throws {SchemaError} If a table with the same name already exists.
   */
  table<RowShape>(
    name: string,
    cols: ColumnShorthandMap<RowShape>,
  ): Table<RowShape> {
    if (this.hasTable(name)) {
      throw new SchemaError("Duplicate table name", name);
    }

    const newTable = new Table(this, name, cols);
    this.tables[name] = newTable;
    return newTable;
  }

  /**
   * Attaches an existing table to the database.
   * @param table The table to attach.
   * @throws {SchemaError} If a table with the same name already exists.
   */
  // biome-ignore lint/suspicious/noExplicitAny: <explanation>
  attachTable(table: Table<any>) {
    if (this.hasTable(table.name)) {
      throw new SchemaError("Duplicate table name", table.name);
    }

    this.tables[table.name] = table;
  }
  /**
   * Executes a function within a database transaction.
   * If the function throws an error, the transaction is rolled back.
   * Otherwise, the transaction is committed.
   * @param transactionFn The function to execute within the transaction.
   * @returns `true` if the transaction was committed successfully, otherwise `false`.
   * @throws The error thrown by the transaction function if the transaction fails.
   */
  transaction<T>(transactionFn: () => T): T {
    try {
      this.db.exec("BEGIN TRANSACTION;");
      const result: T = transactionFn();
      this.db.exec("COMMIT;");
      return result;
    } catch (error) {
      this.db.exec("ROLLBACK;");
      throw error;
    }
  }
}
