import type { Database, SQLQueryBindings, Statement } from "bun:sqlite";
import {
  type ColumnMap,
  type ColumnShorthandMap,
  type ColumnValue,
  type ColumnDataType,
  generateColumnDefinitionSQL,
  normalizeColumns,
} from "./columns";
import { LookupFailure, SchemaError } from "./error";
import { Prequel } from "./prequel";

export interface QueryOptions {
  order?: string;
  limit?: number;
  skip?: number;
}

/**
 * Represents a table in a database.
 *
 * @typeParam RowShape - The shape of the rows in the table.
 */
export class Table<RowShape> {
  private _name: string;
  private _db: Database;
  private _columnDefinitions: ColumnMap<RowShape>;
  private _columnNames: string[];
  private _createTableSQL: string;
  private _primaryColumnName = "ROWID";
  private _primaryColumnType: ColumnDataType | null = null;
  private _insertQuery: Statement<RowShape, [SQLQueryBindings]>;
  private _updateQuery: Statement<RowShape, [SQLQueryBindings]>;
  private _findQuery: Statement<RowShape, [string]>;
  private _deleteQuery: Statement<void, [string]>;
  private _deleteRowQuery: Statement<void, [SQLQueryBindings]>;
  private _sizeQuery: Statement<{ count: number }, []>;
  private _truncateQuery: Statement<void, []>;

  public query: typeof Database.prototype.query;
  public prepare: typeof Database.prototype.prepare;

  constructor(
    db: Database | Prequel,
    name: string,
    cols: ColumnShorthandMap<RowShape>,
  ) {
    if (db instanceof Prequel) {
      this._db = db.db;
    } else {
      this._db = db;
    }
    this._name = name;
    this._columnDefinitions = normalizeColumns(cols);
    this._columnNames = Object.keys(this._columnDefinitions);

    // Pass-thru functions
    this.query = this._db.query.bind(this._db);
    this.prepare = this._db.prepare.bind(this._db);

    const columnSQLParts: string[] = [];
    const updateColumnSQLParts: string[] = [];

    for (const columnName of this._columnNames) {
      const columnDefinition =
        this._columnDefinitions[
          columnName as keyof typeof this._columnDefinitions
        ];

      // Identify a primary column besides ROWID if specified
      if (columnDefinition.primary && this._primaryColumnName === "ROWID") {
        this._primaryColumnName = columnName;
        this._primaryColumnType = columnDefinition.type;
      }

      // Add non-primary columns to the generic update query
      if (!columnDefinition.primary) {
        updateColumnSQLParts.push(`"${columnName}" = $${columnName}`);
      }

      // Generate column definitions for CREATE TABLE
      columnSQLParts.push(
        generateColumnDefinitionSQL(columnName, columnDefinition),
      );
    }

    // Generate SQL for common queries
    const columnSQL = columnSQLParts.join(", ");
    const updateColumnSQL = updateColumnSQLParts.join(", ");
    const allCols = this._columnNames.join(", ");
    const allColVars = this._columnNames
      .map((colName) => `$${colName}`)
      .join(", ");

    // Ensure the table exists in the database
    this._createTableSQL = `CREATE TABLE IF NOT EXISTS "${name}" (${columnSQL});`;
    this.db.exec(this._createTableSQL);

    // Prepare common queries
    this._insertQuery = this.db.query<RowShape, SQLQueryBindings>(
      `INSERT INTO "${name}" (${allCols}) VALUES (${allColVars}) RETURNING *;`,
    );

    this._updateQuery = this.db.query<RowShape, SQLQueryBindings>(
      `UPDATE "${name}" SET ${updateColumnSQL} WHERE "${this._primaryColumnName}" = $${this._primaryColumnName} RETURNING *;`,
    );

    this._findQuery = this.db.query<RowShape, string>(
      `SELECT * FROM "${name}" WHERE "${this._primaryColumnName}" = ? LIMIT 1;`,
    );

    this._deleteQuery = this.db.query<void, string>(
      `DELETE FROM "${name}" WHERE "${this._primaryColumnName}" = ?;`,
    );

    this._sizeQuery = this.db.query<{ count: number }, []>(
      `SELECT count(*) as count FROM ${name}`,
    );

    this._deleteRowQuery = this.db.query<void, SQLQueryBindings>(
      `DELETE FROM "${name}" WHERE "${this._primaryColumnName}" = $${this._primaryColumnName};`,
    );

    this._truncateQuery = this.db.query<void, []>(`DELETE FROM "${name}"`);

    // If using with a Prequel instance, ensure the table is attached
    // This is only for when creating tables directly and passing in a Prequel instance
    if (db instanceof Prequel) {
      if (!db.hasTable(name)) {
        db.attachTable(this);
      }
    }
  }

  protected hasColumn(colName: string) {
    return this._columnNames.includes(colName);
  }

  public primaryColumnType(): ColumnDataType {
    return this._primaryColumnType ?? "INTEGER";
  }

  /**
   * Generates a reference string for a column in the table.
   *
   * @param colName - The name of the column to reference. If not provided, the primary column name is used.
   * @returns A string representing the reference to the specified column or the primary column if no column name is provided.
   * @throws {SchemaError} If the specified column name does not exist in the table schema.
   */
  public reference(colName?: keyof RowShape): string {
    if (colName === undefined) {
      return `${this._name}(${this._primaryColumnName})`;
    }

    const asString = String(colName);

    if (!this.hasColumn(asString)) {
      throw new SchemaError(
        "No such column available for reference",
        this._name,
        asString,
      );
    }

    return `${this._name}(${asString})`;
  }

  /**
   * Inserts a new row into the table.
   *
   * @param data - The data to be inserted, conforming to the RowShape interface.
   * @returns The inserted row, conforming to the RowShape interface.
   */
  public insert(data: RowShape): RowShape {
    return this._insertQuery.get(data as SQLQueryBindings) as RowShape;
  }

  /**
   * Inserts a partial row into the table and returns the inserted row.
   *
   * @param data - An object containing the partial data to be inserted. The keys should match the column names of the table.
   * @returns The inserted row as an object of type `RowShape`.
   *
   * @typeParam RowShape - The shape of the row in the table.
   */
  public insertPartial(data: Partial<RowShape>): RowShape {
    const keys = Object.keys(data);
    const colNames = keys.join(", ");
    const varNames = keys.map((keyName) => `$${keyName}`);

    const query = this._db.prepare(
      `INSERT INTO ${this} (${colNames}) VALUES (${varNames}) RETURNING *;`,
    );

    return query.get(data as SQLQueryBindings) as RowShape;
  }

  public insertWithout<OmitKeys extends keyof RowShape>(
    rowWithKeysOmitted: Omit<RowShape, OmitKeys>,
  ): RowShape {
    return this.insertPartial(rowWithKeysOmitted as Partial<RowShape>);
  }

  /**
   * Updates a row in the table with the provided data.
   *
   * @param data - The data to update the row with, conforming to the RowShape interface.
   * @returns The updated row data if the update is successful, or `null` if it fails.
   * @throws Will throw an error if the primary column name is "ROWID".
   */
  public update(data: RowShape): RowShape | null {
    if (this._primaryColumnName === "ROWID") {
      throw new Error(
        "Cannot use `Table.update()` without setting a primary column",
      );
    }

    return this._updateQuery.get(data as SQLQueryBindings) as RowShape;
  }

  /**
   * Updates rows in the table that match the specified conditions with the provided data.
   *
   * @param data - An object containing the data to update. The keys should correspond to the column names.
   * @param conditions - An object containing the conditions to match the rows that need to be updated. The keys should correspond to the column names.
   * @returns An array of updated rows.
   */
  public updateWhere(
    data: Partial<RowShape>,
    conditions: Partial<RowShape>,
  ): RowShape[] {
    const keys = Object.keys(data);
    const setParts = keys.map((key) => `"${key}" = $${key}`);
    const setClause = setParts.join(", ");

    const whereKeys = Object.keys(conditions);
    const whereParts = whereKeys.map((key) => `"${key}" = $where_${key}`);
    const whereClause = whereParts.join(" AND ");

    const whereData = Object.entries(conditions).reduce(
      (acc, [key, value]) => {
        acc[`where_${key}`] = value as ColumnValue;
        return acc;
      },
      {} as Record<string, ColumnValue>,
    );

    const query = this._db.prepare<RowShape, SQLQueryBindings>(
      `UPDATE ${this} SET ${setClause} WHERE ${whereClause} RETURNING *;`,
    );

    const queryData = {
      ...data,
      ...whereData,
    };

    return query.all(queryData);
  }

  /**
   * Updates all rows in the table with the provided data and returns the updated rows.
   *
   * @param data - An object containing the partial data to update the rows with.
   *               The keys of this object should match the column names of the table.
   * @returns An array of updated rows.
   */
  public updateAll(data: Partial<RowShape>): RowShape[] {
    const keys = Object.keys(data);
    const setParts = keys.map((key) => `"${key}" = $${key}`);
    const setClause = setParts.join(", ");

    const query = this._db.prepare<RowShape, SQLQueryBindings>(
      `UPDATE ${this} SET ${setClause} RETURNING *;`,
    );

    return query.all(data as SQLQueryBindings);
  }

  /**
   * Checks if an entry with the specified ID exists in the table.
   *
   * @param id - The ID of the entry to check for existence.
   * @returns `true` if an entry with the specified ID exists, otherwise `false`.
   */
  public exists(id: ColumnValue): boolean {
    return this.findOneById(id) !== null;
  }

  /**
   * Checks if a row exists in the table that matches the given conditions.
   *
   * @param conditions - An object representing the conditions to match against.
   *                      Each key-value pair represents a column and the value to match.
   * @returns `true` if a matching row is found, otherwise `false`.
   */
  public existsWhere(conditions: Partial<RowShape>): boolean {
    return this.findOneWhere(conditions) !== null;
  }

  /**
   * Counts the number of rows in the table that match the specified conditions.
   * @param conditions The shape of data to count rows by
   * @returns The number of rows that match the specified conditions.
   */
  public countWhere(conditions: Partial<RowShape>): number {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key}`);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare<{ count: number }, SQLQueryBindings>(
      `SELECT count(*) as count FROM ${this} WHERE ${whereClause};`,
    );

    const countResult = query.get(conditions as SQLQueryBindings) ?? {
      count: 0,
    };
    return countResult.count;
  }

  /**
   * Finds a single row by its unique identifier.
   *
   * @param id - The unique identifier of the row to find.
   * @returns The row shape if found, otherwise `null`.
   */
  public findOneById(id: ColumnValue): RowShape | null {
    return this._findQuery.get(`${id}`);
  }

  /**
   * Finds a single row by its unique identifier. Throws an error if no row is found.
   *
   * @param id - The unique identifier of the row to find.
   * @returns The row shape if found, otherwise `null`.
   *
   * @throws {Error} If no row is found with the specified ID.
   */
  public findOneByIdOrFail(id: ColumnValue): RowShape {
    const found = this.findOneById(id);

    if (!found) {
      throw new LookupFailure(this.name, {
        [this._primaryColumnName]: id,
      });
    }

    return found;
  }

  /**
   * Retrieves all rows from the table.
   *
   * @returns {RowShape[]} An array of all rows in the table.
   */
  public findAll(options?: {
    order?: string;
    limit?: number;
    skip?: number;
  }): RowShape[] {
    const optionParts: string[] = [];

    if (options?.order) {
      optionParts.push(`ORDER BY ${options.order}`);
    }

    if (options?.limit) {
      optionParts.push(`LIMIT ${options.limit}`);

      if (options?.skip) {
        optionParts.push(`OFFSET ${options.skip}`);
      }
    }

    return this._db
      .prepare<RowShape, []>(`SELECT * FROM ${this} ${optionParts.join(" ")}`)
      .all();
  }

  /**
   * Finds a single row in the table that matches the specified conditions.
   *
   * @param conditions - An object representing the conditions to match. The keys are column names and the values are the values to match.
   * @returns The first row that matches the conditions, or `null` if no matching row is found.
   */
  public findOneWhere(conditions: Partial<RowShape>): RowShape | null {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key}`);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare<RowShape, SQLQueryBindings>(
      `SELECT * FROM ${this} WHERE ${whereClause}; `,
    );

    return query.get(conditions as SQLQueryBindings);
  }

  /**
   * Finds a single row in the table that matches the specified conditions. Throws an error if no row is found.
   *
   * @param conditions - An object representing the conditions to match. The keys are column names and the values are the values to match.
   * @returns The first row that matches the conditions, or `null` if no matching row is found.
   *
   * @throws {Error} If no row is found that matches the specified conditions.
   */
  public findOneWhereOrFail(conditions: Partial<RowShape>): RowShape {
    const found = this.findOneWhere(conditions);

    if (!found) {
      throw new LookupFailure(this.name, conditions);
    }

    return found;
  }

  /**
   * Finds all rows in the table that match the specified conditions.
   *
   * @param conditions - An object containing key-value pairs that represent the conditions for the query.
   *                     The keys are column names and the values are the values to match.
   * @returns An array of rows that match the specified conditions.
   */
  public findAllWhere(
    conditions: Partial<RowShape>,
    options?: {
      order?: string;
      limit?: number;
      skip?: number;
    },
  ): RowShape[] {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key} `);
    const whereClause = whereParts.join(" AND ");

    const optionParts: string[] = [];

    if (options?.order) {
      optionParts.push(`ORDER BY ${options.order} `);
    }

    if (options?.limit) {
      optionParts.push(`LIMIT ${options.limit} `);

      if (options?.skip) {
        optionParts.push(`OFFSET ${options.skip} `);
      }
    }

    const query = this._db.prepare<RowShape, SQLQueryBindings>(
      `SELECT * FROM ${this} WHERE ${whereClause} ${optionParts.join(" ")}; `,
    );

    return query.all(conditions as SQLQueryBindings);
  }

  /**
   * Deletes a row
   *
   * @param row - The row in the table to delete
   * @returns The updated row data if the update is successful, or `null` if it fails.
   * @throws Will throw an error if the primary column name is "ROWID".
   */
  public delete(data: RowShape): void {
    if (this._primaryColumnName === "ROWID") {
      throw new Error(
        "Cannot use `Table.delete()` without setting a primary column",
      );
    }

    this._deleteRowQuery.run(data as SQLQueryBindings);
  }

  /**
   * Deletes a record from the table by its ID.
   *
   * @param id - The unique identifier of the record to be deleted.
   * @returns void
   */
  public deleteById = (id: ColumnValue): void => {
    if (this._primaryColumnName === "ROWID") {
      throw new Error(
        "Cannot use `Table.deleteById()` without setting a primary column",
      );
    }

    this._deleteQuery.run(`${id} `);
  };

  /**
   * Deletes all rows in the table that match the specified conditions.
   *
   * @param conditions - An object representing the conditions to match. The keys are column names and the values are the values to match.
   */
  public deleteWhere(conditions: Partial<RowShape>): void {
    const keys = Object.keys(conditions);
    const whereParts = keys.map((key) => `"${key}" = $${key} `);
    const whereClause = whereParts.join(" AND ");
    const query = this._db.prepare<void, SQLQueryBindings>(
      `DELETE FROM ${this} WHERE ${whereClause}; `,
    );

    query.run(conditions as SQLQueryBindings);
  }

  /**
   * Deletes all rows in the table.
   */
  public deleteAll(): void {
    this._truncateQuery.run();
  }

  /**
   * Retrieves the size of the table.
   *
   * @returns {number} The number of rows in the table.
   * @throws {Error} If the row count query fails.
   */
  public count(): number {
    const res = this._sizeQuery.get();

    if (res === null) {
      throw new Error("Failed to count rows");
    }

    return res.count;
  }

  public toString() {
    return `"${this._name}"`;
  }

  get name() {
    return this._name;
  }

  get db() {
    return this._db;
  }

  get columns() {
    return this._columnNames;
  }

  get schema(): string {
    return this._createTableSQL;
  }

  get idColumnName(): string {
    return this._primaryColumnName;
  }
}
